/*
 * linear.cpp implements the problem classes and operations related to
 * training/prediction The algorithms are implemented in solver.h/solver.cpp.
 * For handling of large-scale data, we use sparse matrices for the data, while
 * the model is stored as a dense vector
 * */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <locale.h>
#include "linear.h"
#include "solver.h"
typedef signed char schar;
template <class T> static inline void swap(T& x, T& y) { T t=x; x=y; y=t; }
#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif
#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif
template <class S, class T> static inline void clone(T*& dst, S* src, int n)
{
	dst = new T[n];
	memcpy((void *)dst,(void *)src,sizeof(T)*n);
}
#define Malloc(type,n) (type *)malloc(size_t(n)*sizeof(type))
#define INF HUGE_VAL

static void print_string_stdout(const char *s)
{
	fputs(s,stdout);
	fflush(stdout);
}

static void (*liblinear_print_string) (const char *) = &print_string_stdout;

class sparse_operator //Operations for sparse vector/matrix
{
public:
	static double nrm2_sq(const feature_node *x)
	{
		double ret = 0;
		while(x->index != -1)
		{
			ret += x->value*x->value;
			x++;
		}
		return (ret);
	}

	static double dot(const double *s, const feature_node *x)
	{
		double ret = 0;
		while(x->index != -1)
		{
			ret += s[x->index-1]*x->value;
			x++;
		}
		return (ret);
	}

	static void axpy(const double a, const feature_node *x, double *y)
	{
		while(x->index != -1)
		{
			y[x->index-1] += a*x->value;
			x++;
		}
	}
};

class l2r_erm_fun: public function //The abstract empirical loss minimization problem class for shared functionalities
{
	//Note that in the implementation, X is X^T in the paper and vice versa
public:
	l2r_erm_fun(const problem *prob, double *C);
	~l2r_erm_fun();

	double fun(double *w);
	double line_search(double *p, double gTp, double pTp, double wTp, double *fnew, double eta, double init_step_size, double *Xp = NULL);// eta is c_1 in the paper
	int get_nr_variable(void);

	void Xv(double *v, double *Xv);// Compute X^T v for a given vector v
	void Xv(double *v, double *Xv, int vecs); // Compute X^T v for a given matrix v, where vecs is the number of columns of v. v is column major but the output Xv is row major for the ease of later computation
	void get_wx(double *external_wx); // Output X^T w to the array externa_wx
	virtual void get_diagH(double *M) = 0; // Output the diagonal entries of the Hessian to the vector M
	virtual void sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length) = 0; //Return the vector of sqrt(D_w). If it is sparse, **index is the pointer to the index set vector passed by reference, and index_length is the size of the index set as a return value passed by reference. If it is dense, index_length is set to 0.
	virtual void grad_coeff(double *C_times_g) = 0;
	double compute_rho();

protected:
	virtual double C_times_loss(int i, double wx_i) = 0; // The loss function associated with the i-th instance, given x_i^T w, times is weight C
	void XTv(double *v, double *XTv); // XTv += X v, will not overwrite the original value but will add to the original value

	double upper_bound;//An upper bound of the Lipschitz constant of the derivative of the individual loss function. 1 for Squared-hinge loss and 1/4 for logistic loss
	double *C;
	const problem *prob;
	double *wx;
	double *tmp;//A buffer vector used for computing intermediate results for the instances
	double wTw;
	double current_f;
	double current_loss;
};

l2r_erm_fun::l2r_erm_fun(const problem *prob, double *C)
{
	int l = prob->l;
	upper_bound = 1;

	this->prob = prob;

	wx = new double[l];
	tmp = new double[l];
	this->C = C;
}

l2r_erm_fun::~l2r_erm_fun()
{
	delete[] wx;
	delete[] tmp;
}

double l2r_erm_fun::fun(double *w)
{
	int i;
	int l = prob->l;
	int w_size = get_nr_variable();

	wTw = 0;
	current_f = 0;
	Xv(w, wx);

	/* The regularization term */
	double wTw_tmp = 0;
	for(i=0;i<w_size;i++)
		wTw_tmp += w[i]*w[i];
	wTw = wTw_tmp;

	/* The loss term times its weight */
	current_loss = 0;
	for(i=0;i<l;i++)
		current_loss += C_times_loss(i, wx[i]);

	current_f = current_loss + wTw / 2;

	return(current_f);
}

int l2r_erm_fun::get_nr_variable(void)
{
	return prob->n;
}

// Backtracking line search. eta is c_1 in the paper
double l2r_erm_fun::line_search(double *p, double gTp, double pTp, double wTp, double *fnew, double eta, double init_step_size, double *Xp)
{
	int i;
	int l = prob->l;
	const int max_num_linesearch = 100;
	double *tmp_ptr;

	if (Xp == NULL) // If X^T d is available then use it, otherwise compute from scratch and store in tmp
		Xv(p, tmp);
	else
	{
		tmp_ptr = tmp;
		tmp = Xp;
	}

	int num_linesearch;
	double theta = init_step_size;
	if (theta == 1.0)
		for (i=0;i<l;i++)
			wx[i] += tmp[i];// For numerical reasons and better speed to avoid multiplied by 1.0
	else
		for (i=0;i<l;i++)
			wx[i] += theta * tmp[i];// Update X^T w
	for(num_linesearch=0; num_linesearch < max_num_linesearch; num_linesearch++)
	{
		double tmp_fnew = wTw / 2 + theta * wTp + theta * theta * pTp / 2;//\|w + theta p\|^2 = w^T w + 2 theta w^T p + alph^2 p^T p
		for(i=0;i<l;i++)
			tmp_fnew += C_times_loss(i, wx[i]);
		*fnew = tmp_fnew;
		if (*fnew - current_f <= theta * eta * gTp)// The Armijo rule
			break;
		else
		{
			theta *= 0.5;
			for (i=0;i<l;i++)
				wx[i] -= theta * tmp[i]; //Update X^T w according to the new step size
		}
	}
	if (Xp != NULL)
		tmp = tmp_ptr;//Move pointer

	if (num_linesearch >= max_num_linesearch) //If too many backtracking steps, then the line search is deemed failed and a stepsize of 0 is returned, here we roll back the objective and X^T w too
	{
		*fnew = current_f;
		for (i=0;i<l;i++)
			wx[i] -= theta * tmp[i];
		return 0;
	}

	wTw += 2 * theta * wTp + theta * theta * pTp;
	current_f = *fnew;//Update values of objective and w^T w stored in the class
	return theta;
}

void l2r_erm_fun::Xv(double *v, double *Xv)
{
	int i;
	int l = prob->l;
	feature_node **x = prob->x;

	for(i=0;i<l;i++)
		Xv[i] = sparse_operator::dot(v, x[i]);
}

void l2r_erm_fun::Xv(double *v, double *Xv, int vecs)
{
	int i,j;
	int l = prob->l;
	int n = get_nr_variable();
	feature_node **x = prob->x;

	for(i=0;i<l;i++)
		for (j=0;j<vecs;j++)
			Xv[i + j * l] = sparse_operator::dot(v + j * n, x[i]);
}

void l2r_erm_fun::XTv(double *v, double *XTv)//XTv will not be reinitialized, the computed value is added to XTv
{
	int i;
	int l = prob->l;
	feature_node **x = prob->x;

	for(i=0;i<l;i++)
		sparse_operator::axpy(v[i], x[i], XTv);
}

void l2r_erm_fun::get_wx(double *external_wx)
{
	memcpy(external_wx, wx, sizeof(double) * prob->l);
}

double l2r_erm_fun::compute_rho()
{
	int i;
	int l = prob->l;
	double tmp_val = 0;

	// The Frobenius norm squared of the data matrix X times C
	for (i=0; i<l; i++)
	{
		double tmp_val2 = 0;
		feature_node *s = prob->x[i];
		while (s->index!=-1)
		{
			tmp_val2 += s->value * s->value * C[i];
			s++;
		}
		tmp_val += tmp_val2;
	}
	return tmp_val * upper_bound + 1;//See sec 6 of the paper for the Lipschitz constant evaluation formula
}

class l2r_lr_fun: public l2r_erm_fun // Logistic regression
{
public:
	l2r_lr_fun(const problem *prob, double *C);
	~l2r_lr_fun();

	void grad(double *w, double *g);
	void get_diagH(double *M);
	void sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length);
	void grad_coeff(double *C_times_g);

private:
	double *D;// Record the diagonal entries D_w. Retained for L-CommDir-DIAG
	double C_times_loss(int i, double wx_i);
};

l2r_lr_fun::l2r_lr_fun(const problem *prob, double *C):
	l2r_erm_fun(prob, C)
{
	int l = prob->l;
	D = new double[l];
	upper_bound = 0.25;
}

void l2r_lr_fun::get_diagH(double *M)
{
	int i;
	int l = prob->l;
	int w_size=get_nr_variable();
	feature_node **x = prob->x;

	for (i=0; i<w_size; i++) // The part from regularization
		M[i] = 1;

	for (i=0; i<l; i++)
	{
		feature_node *s = x[i];
		while (s->index!=-1)
		{
			M[s->index-1] += s->value * s->value * D[i];
			s++;
		}
	}
}

l2r_lr_fun::~l2r_lr_fun()
{
	delete[] D;
}

double l2r_lr_fun::C_times_loss(int i, double wx_i)
{
	double ywx_i = wx_i * prob->y[i];
	if (ywx_i >= 0)
		return C[i]*log(1 + exp(-ywx_i));
	else
		return C[i]*(-ywx_i + log(1 + exp(ywx_i)));//For better numerical stability, mathematically the same
}

void l2r_lr_fun::grad(double *w, double *g)
{
	int w_size = get_nr_variable();
	grad_coeff(tmp);
	memcpy(g, w, sizeof(double) * w_size);
	XTv(tmp, g);//g += X u_w
}

//Computing u_w
void l2r_lr_fun::grad_coeff(double *C_times_g)
{
	int i;
	double *y = prob->y;
	int l = prob->l;
	double tmp_val;

	for(i=0;i<l;i++)
	{
		tmp_val = 1 / (1 + exp(-y[i] * wx[i]));
		D[i] = C[i] * tmp_val * (1 - tmp_val);
		C_times_g[i] = C[i] * (tmp_val - 1) * y[i];
	}
}


//Square root of D_w
void l2r_lr_fun::sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length)
{
	*index_length = 0;
	for (int i=0;i<prob->l;i++)
		sqrt_C_times_D[i] = sqrt(D[i]);
}

class l2r_l2_svc_fun: public l2r_erm_fun //Squared-hinge loss
{
public:
	l2r_l2_svc_fun(const problem *prob, double *C);
	~l2r_l2_svc_fun();

	void grad(double *w, double *g);

	void sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length);
	void get_diagH(double *M);

protected:
	void grad_coeff(double *C_times_g);
	double C_times_loss(int i, double wx_i);
	void subXTv(double *v, double *XTv);// Compute XTv only for those entries with D_{i,i} != 0

	int *I;//Index set for the active instances
	int sizeI;//Size of I
};

l2r_l2_svc_fun::l2r_l2_svc_fun(const problem *prob, double *C):
	l2r_erm_fun(prob, C)
{
	upper_bound = 2;
	I = new int[prob->l];
}

l2r_l2_svc_fun::~l2r_l2_svc_fun()
{
	delete[] I;
}

double l2r_l2_svc_fun::C_times_loss(int i, double wx_i)
{
	double d = 1 - prob->y[i] * wx_i;
	return (d > 0) * C[i] * d * d;
}

void l2r_l2_svc_fun::grad(double *w, double *g)
{
	int i;
	double *y = prob->y;
	int l = prob->l;
	int w_size = get_nr_variable();
	double tmp_val;

	sizeI = 0;
	for (i=0;i<l;i++)
	{
		tmp_val = wx[i] * y[i] - 1;
		if (tmp_val < 0)//Add one more instance to the index set and store tmp_val to tmp
		{
			tmp[sizeI] = 2 * C[i] * y[i] * tmp_val;
			I[sizeI] = i;
			sizeI++;
		}
	}

	memcpy(g, w, sizeof(double) * w_size);
	subXTv(tmp, g);
}

void l2r_l2_svc_fun::grad_coeff(double *C_times_g)
{
	int i;
	double *y = prob->y;
	int l = prob->l;
	double tmp_val;

	for (i=0;i<l;i++)
	{
		tmp_val = wx[i] * y[i] - 1;
		C_times_g[i] = (tmp_val < 0) * 2 * C[i] * y[i] * tmp_val;
	}
}

void l2r_l2_svc_fun::get_diagH(double *M)
{
	int i;
	int w_size=get_nr_variable();
	int idx;
	feature_node **x = prob->x;

	for (i=0; i<w_size; i++) // The part from regularization
		M[i] = 1;

	for (i=0; i<sizeI; i++)
	{
		idx = I[i];
		feature_node *s = x[idx];
		while (s->index!=-1)
		{
			M[s->index-1] += s->value*s->value * C[idx] * 2;
			s++;
		}
	}
}

void l2r_l2_svc_fun::subXTv(double *v, double *XTv)
{
	int i;
	feature_node **x = prob->x;
	for(i=0;i<sizeI;i++)
		sparse_operator::axpy(v[i], x[I[i]], XTv);
}

void l2r_l2_svc_fun::sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length)
{
	int i;
	*index_length = sizeI;
	for (i=0;i<sizeI;i++)
		sqrt_C_times_D[i] = sqrt(2 * C[I[i]]);
	*index = I;
}

class l2r_l2_svr_fun: public l2r_l2_svc_fun
{
public:
	l2r_l2_svr_fun(const problem *prob, double *C, double p);

	void grad(double *w, double *g);
	void grad_coeff(double *C_times_g);

private:
	double C_times_loss(int i, double wx_i);
	double p;//The epsilon parameter in SVR loss
};

l2r_l2_svr_fun::l2r_l2_svr_fun(const problem *prob, double *C, double p):
	l2r_l2_svc_fun(prob, C)
{
	this->p = p;
}

double l2r_l2_svr_fun::C_times_loss(int i, double wx_i)
{
		double d = wx_i - prob->y[i];
		if(d < -p)
			return C[i]*(d+p)*(d+p);
		else if(d > p)
			return C[i]*(d-p)*(d-p);// Only loss with abs value larger than p counts
		return 0;
}

void l2r_l2_svr_fun::grad(double *w, double *g)
{
	int i;
	double *y = prob->y;
	int l = prob->l;
	int w_size = get_nr_variable();
	double tmp_val;

	sizeI = 0;

	// generate index set I
	for (i=0;i<l;i++)
	{
		tmp_val = wx[i] - y[i];
		if(tmp_val < -p)
		{
			tmp[sizeI] = 2 * C[i] * (tmp_val + p);
			I[sizeI] = i;
			sizeI++;
		}
		else if(tmp_val > p)
		{
			tmp[sizeI] = 2 * C[i] * (tmp_val - p);
			I[sizeI] = i;
			sizeI++;
		}
	}

	memcpy(g, w, sizeof(double) * w_size);
	subXTv(tmp, g);
}

void l2r_l2_svr_fun::grad_coeff(double *C_times_g)
{
	int i;
	double *y = prob->y;
	int l = prob->l;
	double tmp_val;

	for (i=0;i<l;i++)
	{
		tmp_val = wx[i] - y[i];
		if(tmp_val < -p)
			C_times_g[i] = 2 * C[i] * (tmp_val + p);
		else if(tmp_val > p)
			C_times_g[i] = 2 * C[i] * (tmp_val - p);
		else
			C_times_g[i] = 0;
	}
}

/* Start of data pre-processing interface functions */
static void group_classes(const problem *prob, int *nr_class_ret, int **label_ret, int **start_ret, int **count_ret, int *perm)
{
	int l = prob->l;
	int max_nr_class = 16;
	int nr_class = 0;
	int *label = Malloc(int,max_nr_class);
	int *count = Malloc(int,max_nr_class);
	int *data_label = Malloc(int,l);
	int i;

	for(i=0;i<l;i++)
	{
		int this_label = (int)prob->y[i];
		int j;
		for(j=0;j<nr_class;j++)
		{
			if(this_label == label[j])
			{
				++count[j];
				break;
			}
		}
		data_label[i] = j;
		if(j == nr_class)
		{
			if(nr_class == max_nr_class)
			{
				max_nr_class *= 2;
				label = (int *)realloc(label,size_t(max_nr_class)*sizeof(int));
				count = (int *)realloc(count,size_t(max_nr_class)*sizeof(int));
			}
			label[nr_class] = this_label;
			count[nr_class] = 1;
			++nr_class;
		}
	}

	//
	// Labels are ordered by their first occurrence in the training set. 
	// However, for two-class sets with -1/+1 labels and -1 appears first, 
	// we swap labels to ensure that internally the binary SVM has positive data corresponding to the +1 instances.
	//
	if (nr_class == 2 && label[0] == -1 && label[1] == 1)
	{
		swap(label[0],label[1]);
		swap(count[0],count[1]);
		for(i=0;i<l;i++)
		{
			if(data_label[i] == 0)
				data_label[i] = 1;
			else
				data_label[i] = 0;
		}
	}

	int *start = Malloc(int,nr_class);
	start[0] = 0;
	for(i=1;i<nr_class;i++)
		start[i] = start[i-1]+count[i-1];
	for(i=0;i<l;i++)
	{
		perm[start[data_label[i]]] = i;
		++start[data_label[i]];
	}
	start[0] = 0;
	for(i=1;i<nr_class;i++)
		start[i] = start[i-1]+count[i-1];

	*nr_class_ret = nr_class;
	*label_ret = label;
	*start_ret = start;
	*count_ret = count;
	free(data_label);
}

/* Assign solver to optimize the provided problem after data post-processing */
static void train_one(const problem *prob, const parameter *param, double *w, double Cp, double Cn)
{
	//inner and outer tolerances for TRON
	double eps = param->eps;
	int pos = 0;
	int neg = 0;
	for(int i=0;i<prob->l;i++)
		if(prob->y[i] > 0)
			pos++;
	neg = prob->l - pos;
	double primal_solver_tol = eps*max(min(pos,neg), 1)/prob->l;

	function *fun_obj=NULL;

	double *C = new double[prob->l];
	switch(param->problem_type)
	{
		case LOGISTIC:
		{
			for(int i = 0; i < prob->l; i++)
			{
				if(prob->y[i] > 0)
					C[i] = Cp;
				else
					C[i] = Cn;
			}
			fun_obj=new l2r_lr_fun(prob, C);
			break;
		}
		case L2SVM:
		{
			for(int i = 0; i < prob->l; i++)
			{
				if(prob->y[i] > 0)
					C[i] = Cp;
				else
					C[i] = Cn;
			}
			fun_obj=new l2r_l2_svc_fun(prob, C);
			break;
		}
		case L2SVR:
		{
			for(int i = 0; i < prob->l; i++)
				C[i] = param->C;

			fun_obj=new l2r_l2_svr_fun(prob, C, param->p);
			break;
		}
		default:
			fprintf(stderr, "ERROR: unknown problem_type\n");
			break;
	}
	SOLVER *solver_obj;
	int failed = 0;
	switch(param->solver_type)
	{
		case LCOMM_OPT_SOLVER:
		{
			solver_obj = new LCOMM_OPT(fun_obj, prob->l, primal_solver_tol, param->eta, 1, 1.0);
			break;
		}
		case AG_SOLVER:
		{
			solver_obj = new AG(fun_obj, primal_solver_tol, 0);
			break;
		}
		case LCOMM_BFGS_SOLVER:
		{
			solver_obj = new LCOMM_BFGS(fun_obj, prob->l, primal_solver_tol, param->m, param->eta);
			break;
		}
		case LCOMM_DIAG_SOLVER:
		{
			solver_obj = new LCOMM_DIAG(fun_obj, prob->l, primal_solver_tol, param->m, param->eta);
			break;
		}
		default:
		{
			failed = 1;
			fprintf(stderr, "ERROR: unknown solver_type\n");
			break;
		}
	}
	if (failed != 1)
	{
		solver_obj->set_print_string(liblinear_print_string);
		solver_obj->solve(w);
		delete solver_obj;
	}
	delete fun_obj;
	delete[] C;
}

//
// Interface functions
//
// Data preprocessing
model* train(const problem *prob, const parameter *param)
{
	int i;
	int l = prob->l;
	int n = prob->n;
	int w_size = prob->n;
	model *model_ = Malloc(model,1);

	if(prob->bias>=0)
		model_->nr_feature=n-1;
	else
		model_->nr_feature=n;
	model_->param = *param;
	model_->bias = prob->bias;
	if(check_regression_model(model_))
	{
		model_->w = Malloc(double, w_size);
		for(i=0; i<w_size; i++)
			model_->w[i] = 0;
		model_->nr_class = 2;
		model_->label = NULL;
		train_one(prob, param, model_->w, 0, 0);
	}
	else
	{
		int nr_class;
		int *label = NULL;
		int *start = NULL;
		int *count = NULL;
		int *perm = Malloc(int,size_t(l));

		// group training data of the same class
		group_classes(prob,&nr_class,&label,&start,&count,perm);

		model_->nr_class=nr_class;
		model_->label = Malloc(int,size_t(nr_class));
		for(i=0;i<nr_class;i++)
			model_->label[i] = label[i];

		// calculate weighted C
		double *weighted_C = Malloc(double, size_t(nr_class));
		for(i=0;i<nr_class;i++)
			weighted_C[i] = param->C;

		// constructing the subproblem
		feature_node **x = Malloc(feature_node *,size_t(l));
		for(i=0;i<l;i++)
			x[i] = prob->x[perm[i]];

		int k;
		problem sub_prob;
		sub_prob.l = l;
		sub_prob.n = n;
		sub_prob.x = Malloc(feature_node *,sub_prob.l);
		sub_prob.y = Malloc(double,sub_prob.l);

		for(k=0; k<sub_prob.l; k++)
			sub_prob.x[k] = x[k];

		if(nr_class == 2)
		{
			model_->w=Malloc(double, w_size);

			int e0 = start[0]+count[0];
			k=0;
			for(; k<e0; k++)
				sub_prob.y[k] = +1;
			for(; k<sub_prob.l; k++)
				sub_prob.y[k] = -1;

			train_one(&sub_prob, param, &model_->w[0], weighted_C[0], weighted_C[1]);
		}
		else
		{
			model_->w=Malloc(double, w_size*nr_class);
			double *w=Malloc(double, w_size);
			for(i=0;i<nr_class;i++)
			{
				int si = start[i];
				int ei = si+count[i];

				k=0;
				for(; k<si; k++)
					sub_prob.y[k] = -1;
				for(; k<ei; k++)
					sub_prob.y[k] = +1;
				for(; k<sub_prob.l; k++)
					sub_prob.y[k] = -1;

				train_one(&sub_prob, param, w, weighted_C[i], param->C);

				for(int j=0;j<w_size;j++)
					model_->w[j*nr_class+i] = w[j];
			}
			free(w);
		}

		free(x);
		free(label);
		free(start);
		free(count);
		free(perm);
		free(sub_prob.x);
		free(sub_prob.y);
		free(weighted_C);
	}
	return model_;
}

/* The decision value before converting to the decision label */
double predict_values(const struct model *model_, const struct feature_node *x, double *dec_values)
{
	int idx;
	int n;
	if(model_->bias>=0)
		n=model_->nr_feature+1;
	else
		n=model_->nr_feature;
	double *w=model_->w;
	int nr_class=model_->nr_class;
	int i;
	int nr_w;
	if(nr_class==2)
		nr_w = 1;
	else
		nr_w = nr_class;

	const feature_node *lx=x;
	for(i=0;i<nr_w;i++)
		dec_values[i] = 0;
	for(; (idx=lx->index)!=-1; lx++)
	{
		// the dimension of testing data may exceed that of training
		if(idx<=n)
			for(i=0;i<nr_w;i++)
				dec_values[i] += w[(idx-1)*nr_w+i]*lx->value;
	}

	if(nr_class==2)
	{
		if(check_regression_model(model_))
			return dec_values[0];
		else
			return (dec_values[0]>0)?model_->label[0]:model_->label[1];
	}
	else
	{
		int dec_max_idx = 0;
		for(i=1;i<nr_class;i++)
		{
			if(dec_values[i] > dec_values[dec_max_idx])
				dec_max_idx = i;
		}
		return model_->label[dec_max_idx];
	}
}

/* Predicted label */
double predict(const model *model_, const feature_node *x)
{
	double *dec_values = Malloc(double, model_->nr_class);
	double label = predict_values(model_, x, dec_values);
	free(dec_values);
	return label;
}


static const char *problem_type_table[]=
{
	"L2R_LR","L2R_L2SVM", "L2R_L2SVR", NULL
};

int save_model(const char *model_file_name, const struct model *model_)
{
	int i;
	int nr_feature=model_->nr_feature;
	int n;
	const parameter& param = model_->param;

	if(model_->bias>=0)
		n=nr_feature+1;
	else
		n=nr_feature;
	int w_size = n;
	FILE *fp = fopen(model_file_name,"w");
	if(fp==NULL) return -1;

	char *old_locale = setlocale(LC_ALL, NULL);
	if (old_locale)
	{
		old_locale = strdup(old_locale);
	}
	setlocale(LC_ALL, "C");

	int nr_w;
	if(model_->nr_class==2)
		nr_w=1;
	else
		nr_w=model_->nr_class;

	fprintf(fp, "problem_type %s\n", problem_type_table[param.problem_type]);
	fprintf(fp, "nr_class %d\n", model_->nr_class);

	if(model_->label)
	{
		fprintf(fp, "label");
		for(i=0; i<model_->nr_class; i++)
			fprintf(fp, " %d", model_->label[i]);
		fprintf(fp, "\n");
	}

	fprintf(fp, "nr_feature %d\n", nr_feature);

	fprintf(fp, "bias %.16g\n", model_->bias);

	fprintf(fp, "w\n");
	for(i=0; i<w_size; i++)
	{
		int j;
		for(j=0; j<nr_w; j++)
			fprintf(fp, "%.16g ", model_->w[i*nr_w+j]);
		fprintf(fp, "\n");
	}

	setlocale(LC_ALL, old_locale);
	free(old_locale);

	if (ferror(fp) != 0 || fclose(fp) != 0) return -1;
	else return 0;
}

//
// FSCANF helps to handle fscanf failures.
// Its do-while block avoids the ambiguity when
// if (...)
//    FSCANF();
// is used
//
#define FSCANF(_stream, _format, _var)do\
{\
	if (fscanf(_stream, _format, _var) != 1)\
	{\
		fprintf(stderr, "ERROR: fscanf failed to read the model\n");\
		EXIT_LOAD_MODEL()\
	}\
}while(0)
// EXIT_LOAD_MODEL should NOT end with a semicolon.
#define EXIT_LOAD_MODEL()\
{\
	setlocale(LC_ALL, old_locale);\
	free(model_->label);\
	free(model_);\
	free(old_locale);\
	return NULL;\
}
struct model *load_model(const char *model_file_name)
{
	FILE *fp = fopen(model_file_name,"r");
	if(fp==NULL) return NULL;

	int i;
	int nr_feature;
	int n;
	int nr_class;
	double bias;
	model *model_ = Malloc(model,1);
	parameter& param = model_->param;

	model_->label = NULL;

	char *old_locale = setlocale(LC_ALL, NULL);
	if (old_locale)
	{
		old_locale = strdup(old_locale);
	}
	setlocale(LC_ALL, "C");

	char cmd[81];
	while(1)
	{
		FSCANF(fp,"%80s",cmd);
		if(strcmp(cmd,"problem_type")==0)
		{
			FSCANF(fp,"%80s",cmd);
			int i;
			for(i=0;problem_type_table[i];i++)
			{
				if(strcmp(problem_type_table[i],cmd)==0)
				{
					param.problem_type=i;
					break;
				}
			}
			if(problem_type_table[i] == NULL)
			{
				fprintf(stderr,"unknown problem type.\n");
				EXIT_LOAD_MODEL()
			}
		}
		else if(strcmp(cmd,"nr_class")==0)
		{
			FSCANF(fp,"%d",&nr_class);
			model_->nr_class=nr_class;
		}
		else if(strcmp(cmd,"nr_feature")==0)
		{
			FSCANF(fp,"%d",&nr_feature);
			model_->nr_feature=nr_feature;
		}
		else if(strcmp(cmd,"bias")==0)
		{
			FSCANF(fp,"%lf",&bias);
			model_->bias=bias;
		}
		else if(strcmp(cmd,"w")==0)
		{
			break;
		}
		else if(strcmp(cmd,"label")==0)
		{
			int nr_class = model_->nr_class;
			model_->label = Malloc(int,nr_class);
			for(int i=0;i<nr_class;i++)
				FSCANF(fp,"%d",&model_->label[i]);
		}
		else
		{
			fprintf(stderr,"unknown text in model file: [%s]\n",cmd);
			EXIT_LOAD_MODEL()
		}
	}

	nr_feature=model_->nr_feature;
	if(model_->bias>=0)
		n=nr_feature+1;
	else
		n=nr_feature;
	int w_size = n;
	int nr_w;
	if(nr_class==2)
		nr_w = 1;
	else
		nr_w = nr_class;

	model_->w=Malloc(double, w_size*nr_w);
	for(i=0; i<w_size; i++)
	{
		int j;
		for(j=0; j<nr_w; j++)
			FSCANF(fp, "%lf ", &model_->w[i*nr_w+j]);
		if (fscanf(fp, "\n") !=0)
		{
			fprintf(stderr, "ERROR: fscanf failed to read the model\n");
			EXIT_LOAD_MODEL()
		}
	}

	setlocale(LC_ALL, old_locale);
	free(old_locale);

	if (ferror(fp) != 0 || fclose(fp) != 0) return NULL;

	return model_;
}

int get_nr_feature(const model *model_)
{
	return model_->nr_feature;
}

int get_nr_class(const model *model_)
{
	return model_->nr_class;
}

void get_labels(const model *model_, int* label)
{
	if (model_->label != NULL)
		for(int i=0;i<model_->nr_class;i++)
			label[i] = model_->label[i];
}

// use inline here for better performance (around 20% faster than the non-inline one)
static inline double get_w_value(const struct model *model_, int idx, int label_idx)
{
	int nr_class = model_->nr_class;
	const double *w = model_->w;

	if(idx < 0 || idx > model_->nr_feature)
		return 0;
	if(check_regression_model(model_))
		return w[idx];
	else
	{
		if(label_idx < 0 || label_idx >= nr_class)
			return 0;
		if(nr_class == 2)
		{
			if(label_idx == 0)
				return w[idx];
			else
				return -w[idx];
		}
		else
			return w[idx*nr_class+label_idx];
	}
}

// feat_idx: starting from 1 to nr_feature
// label_idx: starting from 0 to nr_class-1 for classification models;
//            for regression models, label_idx is ignored.
double get_decfun_coef(const struct model *model_, int feat_idx, int label_idx)
{
	if(feat_idx > model_->nr_feature)
		return 0;
	return get_w_value(model_, feat_idx-1, label_idx);
}

double get_decfun_bias(const struct model *model_, int label_idx)
{
	int bias_idx = model_->nr_feature;
	double bias = model_->bias;
	if(bias <= 0)
		return 0;
	else
		return bias*get_w_value(model_, bias_idx, label_idx);
}

void free_model_content(struct model *model_ptr)
{
	if(model_ptr->w != NULL)
		free(model_ptr->w);
	if(model_ptr->label != NULL)
		free(model_ptr->label);
}

void free_and_destroy_model(struct model **model_ptr_ptr)
{
	struct model *model_ptr = *model_ptr_ptr;
	if(model_ptr != NULL)
	{
		free_model_content(model_ptr);
		free(model_ptr);
	}
}

void destroy_param(parameter* param)
{
	return;
}

const char *check_parameter(const problem *prob, const parameter *param)
{
	if(param->eps <= 0)
		return "eps <= 0";

	if(param->C <= 0)
		return "C <= 0";

	if(param->p < 0)
		return "p < 0";

	if(param->solver_type != LCOMM_OPT_SOLVER && param->solver_type != LCOMM_BFGS_SOLVER && param->solver_type != LCOMM_DIAG_SOLVER && param->solver_type != AG_SOLVER)
		return "unknown solver type";

	if (param->problem_type != LOGISTIC && param->problem_type != L2SVM && param->problem_type != L2SVR)
		return "unknown loss function";
	return NULL;
}

void set_print_string_function(void (*print_func)(const char*))
{
	if (print_func == NULL)
		liblinear_print_string = &print_string_stdout;
	else
		liblinear_print_string = print_func;
}

int check_regression_model(const struct model *model_)
{
	return (model_->param.problem_type==L2SVR);
}
