/*
 * solver.cpp implements the algorithms for comparison
 * In this file, we implement LCOMM_OPT, LCOMM_BFGS, LCOMM_DIAG, and Accelerated gradient
 * */
#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <limits.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "solver.h"

template <class T> static inline void swap(T* a, int x, int y) { T t=a[x]; a[x]=a[y]; a[y]=t; }
#ifndef TIMER
#define TIMER
#define NS_PER_SEC 1000000000
double wall_time_diff(int64_t ed, int64_t st)
{
	return (double)(ed-st)/(double)NS_PER_SEC;
}

int64_t wall_clock_ns()
{
#if __unix__
	struct timespec tspec;
	clock_gettime(CLOCK_MONOTONIC, &tspec);
	return tspec.tv_sec*NS_PER_SEC + tspec.tv_nsec;
#else
#if __MACH__
	return 0;
#else
	struct timeval tv;
	gettimeofday( &tv, NULL );
	return tv.tv_sec*NS_PER_SEC + tv.tv_usec*1000;
#endif
#endif
}
#endif

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#define MEPS 1e-16L
#define PIVOTING
#define DEBUG_LDL

// solve Ax=b , in which A[i*n+j] = Aij and we only solve submatrix A[m:m]
// This subrouine implements the LDLT algorithm with pivoting
double solve_system(const double *_A, const double *_b, int n, int m, double *x)
{
	long double *A = new long double[n * n];
	for(int i=0; i<n*n; i++)
		A[i] = _A[i];
	long double *b = new long double[n];
	for(int i=0; i<n; i++)
		b[i] = _b[i];

#ifdef PIVOTING
	int *p = new int[m];
	for(int i=0; i<m; i++)
		p[i] = i;
#endif

	// decomp
	// A -> D0
	//      L00 D1
	//      L10 L11 D2
	for(int j=0; j<m; j++){
#ifdef PIVOTING
		int pj = j;
		long double pj_val = fabsl(A[j*n+j]);
		// find piv
		for(int i=j+1; i<m; i++){
			long double pi_val = fabsl(A[i*n+i]);
			if(pj_val < pi_val){
				pj_val = pi_val;
				pj = i;
			}
		}
		// swap rows and columns of j and pj
		for(int i=0; i<m; i++)
			swap(A, j*n+i, pj*n+i);
		for(int i=0; i<m; i++)
			swap(A, i*n+j, i*n+pj);
		swap(b, j, pj);
		swap(p, j, pj);
#endif

		// loop invariant: Aik = Lik, Akk = Dk with k<j
		long double Dj = A[j*n+j];
		for(int k=0; k<j; k++){
			Dj -= A[j*n+k]*A[j*n+k] * A[k*n+k];
		}
		A[j*n+j] = Dj;
		//assert(fabs(Dj) >= MEPS);
		if(fabsl(Dj) < MEPS){
			for(int i=j+1; i<m; i++)
				A[i*n+j] = 0;
			continue; // xj should be 0
		}

		for(int i=j+1; i<m; i++){
			long double Lij = A[i*n+j];
			for(int k=0; k<j; k++){
				Lij -= A[i*n+k] * A[j*n+k] * A[k*n+k];
			}
			Lij /= Dj;
			A[i*n+j] = Lij;
		}
	}
#if 0
	printf("L\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<=i; j++)
			printf("%+.3Le ", A[i*n+j]);
		printf("\n");
	}
#endif
	// invariant: A[i*n+j] = Lij, forall i>j
	//            A[i*n+i] = Dii, forall i

	// LD y = b
	// di yi =  bi - \sum_{j<i} yj Lij
	// place y in b
	for(int i=0; i<m; i++){
		// loop invariant: bk = yk forall k<i
		long double yi = b[i];
		for(int k=0; k<i; k++)
			yi -= A[i*n+k] * b[k];
		b[i] = yi;
	}
	for(int i=0; i<m; i++){
		long double Di = A[i*n+i];
		if(fabsl(Di) < MEPS)
			b[i] = 0;
		else
			b[i] /= Di;
	}
#if 0
	printf("y\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", b[i]);
	}
#endif
	// L'x = y
	// xi = yi - \sum{j>i} xj Lij
	for(int i=m-1; i>=0; i--){
		// loop invariant: xk ready forall k>i
		long double xi = b[i];
#ifdef PIVOTING
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[p[k]];
		x[p[i]] = (double)xi;
#else
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[k];
		x[i] = (double)xi;
#endif
	}
#if 0
	printf("\n");
	for(int i=0; i<n; i++)
		printf("d %.3e\n", x[i]);

	printf("A\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<n; j++)
			printf("%+.3e ", Acopy[i*n+j]);
		printf("\n");
	}
	printf("b\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", bcopy[i]);
	}
#endif

#if 1
	double err = 0;
	for(int i=0; i<m; i++){
		double Ax_i = 0;
		for(int j=0; j<m; j++)
			Ax_i += _A[i*n+j] * x[j];
		err += fabs(Ax_i-_b[i]);
	}
	double max_pv = 0;
	for(int i=0; i<m; i++)
		if(max_pv < _A[i*n+i])
			max_pv = (double)_A[i*n+i];
	if(max_pv != 0)
		err /= max_pv;
	//printf("err=%e\n", err);
	//assert(err < 1e-5);
#endif
	delete[] A;
	delete[] b;
#ifdef PIVOTING
	delete[] p;
#endif
	return err;
}

#ifdef __cplusplus
extern "C" {
#endif

extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);
extern int dgemv_(char *, int *, int *, double *, double *, int *, double *, int *, double *, double *, int *);
extern int dsyrk_(char *, char *, int *, int *, double *, double *, int *, double *, double *, int *);
extern int dspmv_(char *, int *, double *, double *, double *, int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}

void SOLVER::info(const char *fmt,...) // Print message to the specified output (either stdout or not to print)
{
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*solver_print_string)(buf);
}

void SOLVER::set_print_string(void (*print_string) (const char *buf))
{
	solver_print_string = print_string;
}

SOLVER::SOLVER(const function *fun_obj, double eps, int max_iter)
{
	this->fun_obj = const_cast<function *>(fun_obj);
	this->eps = eps;
	this->max_iter = max_iter;
	solver_print_string = default_print;
}


LCOMM_OPT::LCOMM_OPT(const function *fun_obj, int l, double eps, double eta, int multi_iter, double sigma, int max_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->l = l;
	this->eta = eta;
	this->multi_iter = multi_iter;
	this->sigma = sigma;//sigma is the strong convexity parameter
	Plength = 3;// P = [w_k, grad_k, v_k]. Plength is the largest possible number of columns in P
	sqrtD_XTP = NULL;
}

LCOMM_OPT::~LCOMM_OPT()
{
	if (sqrtD_XTP != NULL)
		delete[] sqrtD_XTP;
}

void LCOMM_OPT::solve(double *w)
{
	// constants used in BLAS
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int inc = 1;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";

	double tt = 0, innertt = 0, check_tt = 0;// Timers
	int i;
	int stopping = 0; // 1 when the conditions in (21) are satisfied
	int inner_iters = 0;
	int *index = NULL; //Index set, used when D_w and u_w are sparse
	int index_length = 0;//the size of the index set. Set to 0 when the vectors are dense

	sqrtD_XTP = new double[Plength * l];// A temp matrix for storing sqrt(D_w)X^T  P (for computing P^T X D_w X^T P using syrk)
	const int system_size = Plength;//the linear system is at most Plength * Plength

	n = fun_obj->get_nr_variable();


	double f;
	int iter = 0;
	double *P = new double[Plength * n];
	wcounter = 0;//the column of P that stores w
	gcounter = 1;//The column of P that stores the gradient
	vcounter = 2;//The column of P that stores the vector v for achieving optimal rates
	double gnorm;
	double gnorm_at_0;//For computing the relative stopping condition referenced at the point 0
	double gnorm2;//gnorm^2
	double step_size = 0;
	double gTp = 0;
	double err = 0;//Residual (in l1 norm) from the linear system solver
	double pTp;
	double rho = 0;// Lipschitz constant
	
	double* linear = new double[Plength];//The rhs of the linear system
	double* pTptmp = new double[Plength];//A temp vector for computing p^T p using the inner product matrix
	double *sqrt_C_times_D = new double[l];//sqrt(D_w)
	double *inner_product_matrix = new double[(Plength) * (Plength + 1) / 2];//symmetric matrix for storing P^T P, so only need to store the upper triangular part
	double *step = new double[n];//This is p
	double *step_x = new double[l];//This is X^T p
	double* t = new double[Plength];// Stores the current solution to the linear system
	double* accumulated_t = new double[Plength];// The current iterate for t_k
	// Each column represents a possible direction to update, stored in a 1d array but treated as an n * length matrix as in blas
	double *XTP = new double[Plength * l];//X^T P
	double *system = new double[Plength * Plength];//The coefficients of the lhs of the linear system
	double *gcoeff = new double[l];//u_w
	double *vnew = new double[n];// v_{k+1}
	double *XTvnew = new double[l];//X^T v_{k+1}

	double *localw = P + wcounter * n;// The latest iterate (w is the final output)
	double *wx = XTP + wcounter * l; // X^T w
	double *g = P + gcounter * n;//gradient
	double *gx = XTP + gcounter * l;// X^T g
	double *v = P + vcounter * n;// The current v
	double *vx = XTP + vcounter * l; //X^T v
	double outer_f;//Objective at the outer iterations
	rho = fun_obj->compute_rho();

	//Coefficients from eq. (7), with alpha = sqrt(sigma / rho) from eq. (10)
	v_coeff[0] = sqrt(sigma / rho);
	v_coeff[1] = 1 - v_coeff[0];
	v_coeff[2] = - v_coeff[0] / sigma;
	
	memset(localw, 0, sizeof(double) * n);
	memset(wx, 0, sizeof(double) * l);
	memset(w, 0, sizeof(double) * n);
	memcpy(v, localw, sizeof(double) * n);
	memcpy(vx, wx, sizeof(double) * l);
	int64_t timer_st = wall_clock_ns();//Timer start
	int64_t timer2_st, timer2_ed;//Other timers

	//Initialization
	f = fun_obj->fun(localw);
	outer_f = f;
	fun_obj->grad(w, g);
	fun_obj->Xv(g, gx);// Update XTP
	gnorm = dnrm2_(&n, g, &inc);
	gnorm_at_0 = gnorm;
	double tmp = ddot_(&n, w, &inc, w, &inc);//tmp = w^T w

	//Initialize the inner products for P^T P, only update the upper triangular part
	//w_0 = v_0
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp;
	inner_product_matrix[inner_product_entry(wcounter, vcounter)] = tmp;
	inner_product_matrix[inner_product_entry(vcounter, vcounter)] = tmp;
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = ddot_(&n, w, &inc, g, &inc);
	inner_product_matrix[inner_product_entry(vcounter, gcounter)] = inner_product_matrix[inner_product_entry(wcounter, gcounter)];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;

	//Start iterating
	while (iter < max_iter)
	{
		int64_t timer_ed = wall_clock_ns();
		tt = wall_time_diff(timer_ed, timer_st) - check_tt;

		gnorm2 = inner_product_matrix[inner_product_entry(gcounter, gcounter)];//the gradient norm squared is an entry in the inner product matrix
		gnorm = sqrt(gnorm2);

		//compute v_{k+1} and X^T v_{k+1}
		dgemv_(N, &n, &Plength, &one, P, &n, v_coeff, &inc, &zero, vnew, &inc);//Assign v_coeff^T P to vnew
		dgemv_(N, &l, &Plength, &one, XTP, &l, v_coeff, &inc, &zero, XTvnew, &inc);//Assign v_coeff^T (X^T P) to XTvnew

		info("iter %d |g| %5.3e f %15.20e inner_iter %d time %5.3e\n", iter, gnorm, outer_f, inner_iters, tt);
		stopping = 0; //stopping = 1 when conditions in eq. (21) are satisfied, then terminate inner iterations
		inner_iters = 0;
		memset(accumulated_t, 0, sizeof(double) * Plength);
		accumulated_t[0] = 1;//The starting point is G(t) = w in eq. (2) as w is in the range of G
		if (gnorm <= eps*gnorm_at_0)
			break;


		while(stopping == 0)
		{
			timer2_st = wall_clock_ns();
			fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);//Obtain sqrt(D_w)
			if (index_length == 0) //Dense D_w
				compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system);//Compute the matrix in the linear system
			else //Sparse D_w
				compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, index, index_length);
			memset(t, 0, sizeof(double) * Plength);

			//Now compute the rhs vector in the linear system
			if (inner_iters == 0)
				for (i=0;i<Plength;i++)
					linear[i] = -inner_product_matrix[inner_product_entry(i,gcounter)];//The rhs is -P^T grad f when w is not updated
			else // Otherwise, the linear term is updated by updating u_w only without recomputing the gradient vector
			{
				dgemv_(T, &l, &Plength, &minusone, XTP, &l, gcoeff, &inc, &zero, linear, &inc);
				dgemv_(T, &n, &Plength, &minusone, P, &n, w, &inc, &one, linear, &inc);
			}

			err = solve_system(system, linear, Plength, system_size, t);//t = system \ linear

			gTp = -ddot_(&Plength, linear, &inc, t, &inc);//gTp = gT (Pt), linear = -gTP

			if (gTp > 0)// If the update direction is nondescent: reverse the direction
			{
				gTp = -gTp;
				for (i=0;i<Plength;i++)
					t[i] = -t[i];
			}

			//Compute p = Pt and X^T p = (X^T P) t
			dgemv_(N, &n, &Plength, &one, P, &n, t, &inc, &zero, step, &inc);
			dgemv_(N, &l, &Plength, &one, XTP, &l, t, &inc, &zero, step_x, &inc);

			memset(pTptmp, 0, sizeof(double) * Plength);//pTp = (Pt)^T (Pt) = t^T (P^T P) t. Compute using the inner product matrix
			dspmv_(U, &Plength, &one, inner_product_matrix, t, &inc, &zero, pTptmp, &inc);//pTptmp = P^T P t
			pTp = ddot_(&Plength, t, &inc, pTptmp, &inc);
			double wTp = ddot_(&Plength, accumulated_t, &inc, pTptmp, &inc);
			double oldf = f;
			step_size = fun_obj->line_search(step, gTp, pTp, wTp, &f, eta, 1.0, step_x);
			//Error handling
			if (step_size == 0)
			{
				info("WARNING: step size too small\n");
				break;
			}
			double actred = oldf - f;
			if (fabs(actred) < 0)
			{
				info("WARNING: actred < 0\n");
				break;
			}
			if (fabs(actred) <= 1.0e-32*fabs(f))
			{
				info("WARNING: actred too small\n");
				break;
			}

			dscal_(&n, &step_size, step, &inc);//step *= step_size
			dscal_(&l, &step_size, step_x, &inc);// step_x *= step_size
			daxpy_(&n, &one, step, &inc, w, &inc);// w += step
			daxpy_(&Plength, &step_size, t, &inc, accumulated_t, &inc);// accumulated_t += t, to update the inner iterate

			fun_obj->grad_coeff(gcoeff);//Calculate the new v_w for checking the stopping condition

			timer2_ed = wall_clock_ns();
			innertt += wall_time_diff(timer2_ed, timer2_st);
			if (multi_iter == 0)//When multiple inner iterations is turned  off
				stopping = 1;
			else
			{
				timer2_st = wall_clock_ns();
				stopping = check_stopping(XTP, gcoeff, accumulated_t, inner_product_matrix) && (f <= outer_f - gnorm2 / (2 * rho)); //the second condition is (21b)
				timer2_ed = wall_clock_ns();
				check_tt += wall_time_diff(timer2_ed, timer2_st);
			}

			inner_iters++;
			if (multi_iter)
				info("\tInner_iter %d f %15.20e step_size %5.3e err %5.3e Inner_time %5.3e\n", inner_iters, f, step_size, err, innertt);
		}
		if (inner_iters == 0)
			break;
		iter++;
		outer_f = f;

		memcpy(localw, w, sizeof(double) * n);//Update the outer iterate
		fun_obj->get_wx(wx);
		memcpy(v, vnew, sizeof(double) * n);//Update v
		memcpy(vx, XTvnew, sizeof(double) * l);

		fun_obj->grad(localw, g);
		fun_obj->Xv(g, gx);
		update_inner_products(inner_product_matrix, P, accumulated_t);// Update the inner products after updating P
	}

	delete[] pTptmp;
	delete[] t;
	delete[] accumulated_t;
	delete[] system;
	delete[] XTP;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
	delete[] gcoeff;
	delete[] vnew;
	delete[] XTvnew;
}

int LCOMM_OPT::check_stopping(double *XTP, double *gcoeff, double *accumulated_t, double *inner_product_matrix)
{
	//BLAS constants
	char U[] = "U";
	char T[] = "T";
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0;
	int inc = 1;


	double *tmp = new double[Plength];
	double *tmp2 = new double[Plength];

	dgemv_(T, &l, &Plength, &one, XTP, &l, gcoeff, &inc, &zero, tmp, &inc);//tmp = (XTP)^T g_coeff
	memcpy(tmp2, v_coeff, sizeof(double) * Plength);
	daxpy_(&Plength, &minusone, accumulated_t, &inc, tmp2, &inc); //P tmp2 = v_{k+1} - w_{k+1}
	double cond = ddot_(&Plength, tmp, &inc, tmp2, &inc);//g = w + X * g_coeff  ==> g^T y = w^T y + g_coeff^T * (yX)^T for any y, here only computed the 2nd term
	dspmv_(U, &Plength, &one, inner_product_matrix, tmp2, &inc, &zero, tmp, &inc);
	//tmp = P^T P tmp2
	cond += ddot_(&Plength, tmp, &inc, tmp2, &inc) * sigma / 2.0 + ddot_(&Plength, tmp, &inc, accumulated_t, &inc);//(21a)

	delete[] tmp;
	delete[] tmp2;
	return (cond >= 0);
}

void LCOMM_OPT::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system)
{
	double one = 1.0;
	char U[] = "U";
	char T[] = "T";

	int i,j;
	memset(system, 0, sizeof(double) * Plength * Plength);

	for (i=0;i<Plength;i++)
	{
		int counteri = i * l;
		//Fill in the entries of sqrt(D) X^T P corresponds to the i-th column of X^T P
		for (j=0;j<l;j++)
			sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[counteri + j];
		//Compute the part of the linear system from P^T P corresponds to the i-th column
		for (j=0;j<=i;j++)
			system[i * Plength + j] = inner_product_matrix[inner_product_entry(i,j)];
	}

	dsyrk_(U, T, &Plength, &l, &one, sqrtD_XTP, &l, &one, system, &Plength);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part
	for (i=0;i<Plength;i++)
		for (j=i+1;j<Plength;j++)
			system[i * Plength + j] = system[j * Plength + i];//Update the lower triangular part by copy-pasting
}

//When D_w is sparse
void LCOMM_OPT::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int *index, int index_length)
{
	double one = 1.0;
	char U[] = "U";
	char T[] = "T";

	int i,j;
	memset(system, 0, sizeof(double) * Plength * Plength);

	for (i=0;i<Plength;i++)
	{
		int counteri = i * index_length;
		int tmp_counter = i * l;

		//Fill in the entries of sqrt(D) X^T P corresponds to the i-th column of X^T P, ignore columns from X^T P corresponding to zero D_w entries
		for (j=0;j<index_length;j++)
			sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + index[j]];

		//Compute the part of the linear system from P^T P corresponds to the i-th column
		for (j=0;j<=i;j++)
			system[i * Plength + j] = inner_product_matrix[inner_product_entry(i,j)];
	}

	dsyrk_(U, T, &Plength, &index_length, &one, sqrtD_XTP, &index_length, &one, system, &Plength);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part, and note that sqrtD_XTP has only index_length rows
	for (i=0;i<Plength;i++)
		for (j=i+1;j<Plength;j++)
			system[i * Plength + j] = system[j * Plength + i];
}

inline int LCOMM_OPT::inner_product_entry(int i, int j)//If i < j then access the (j,i) entry instead as P^T P is triangular
{
	int index1, index2;
	index1 = (i > j) * i + (i <= j) * j;
	index2 = i + j - index1;
	return index1 * (index1 + 1) / 2 + index2;
}

void LCOMM_OPT::update_inner_products(double* inner_product_matrix, double* P, double* t)
{
	int i;
	int inc = 1;
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";

	double *PTg = new double[Plength];
	double *PTv = new double[Plength];
	double *PTw = new double[Plength];

	dspmv_(U, &Plength, &one, inner_product_matrix, v_coeff, &inc, &zero, PTv, &inc);//PTv = P^T P v_coeff
	dspmv_(U, &Plength, &one, inner_product_matrix, t, &inc, &zero, PTw, &inc);//PTw = P^T Pt

	dgemv_(T, &n, &Plength, &one, P, &n, P + gcounter * n, &inc, &zero, PTg, &inc);//PTg = P^T g

	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = ddot_(&Plength, PTw, &inc, t, &inc);//w^T w = (Pt)^T Pt
	inner_product_matrix[inner_product_entry(wcounter, vcounter)] = ddot_(&Plength, PTw, &inc, v_coeff, &inc);// w^T v = (Pt)^T (P v_coeff)
	inner_product_matrix[inner_product_entry(vcounter, vcounter)] = ddot_(&Plength, PTv, &inc, v_coeff, &inc);// v^T v = |P v_coeff|^2
	for (i=0;i<Plength;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];
	
	delete[] PTg;
	delete[] PTv;
	delete[] PTw;
}


LCOMM_BFGS::LCOMM_BFGS(const function *fun_obj, int l, double eps, int m, double eta, int max_iter): LCOMM_OPT(fun_obj, l, eps, eta, 0, max_iter)
{
	this->M = m;//M is #historical iterations
	Plength = 2 * M + 2;//2 vectors per iteration in P
	this->n = this->fun_obj->get_nr_variable();
}

LCOMM_BFGS::~LCOMM_BFGS()
{
}

//Algorithm 3 in the paper. Here A = I and b = 0, so many operations are simplified
void LCOMM_BFGS::solve(double *w)
{
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int inc = 1;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";

	double tt = 0;//Timer
	int i;
	int *index = NULL; //Index set, used when D_w and u_w are sparse
	int index_length = 0;//the size of the index set. Set to 0 when the vectors are dense
	double f;
	int iter = 0;

	int DynamicM = 0;//min(M,iter)
	int previous_M = 0;//DynamicM at last iteration
	double gnorm;
	double step_size = 0;
	double gTp = 0;
	double err = 0;//Residual (in l1 norm) from the linear system solver
	double pTp;
	double gnorm_at_0;

	sqrtD_XTP = new double[Plength * l];

	wcounter = 0;//the column of P that stores w
	gcounter = 1;//The column of P that stores the gradient
	
	double* linear = new double[Plength];//The rhs of the linear system
	double* pTptmp = new double[Plength];//A temp vector for computing p^T p using the inner product matrix
	double *sqrt_C_times_D = new double[l];//sqrt(D_w)
	double *inner_product_matrix = new double[Plength * Plength];//symmetric matrix, so only need to store the upper triangular part
	double *step = new double[n];//This is p
	double *step_x = new double[l];//This is X^T p
	double* t = new double[Plength];// Stores the current solution to the linear system
	double *P = new double[Plength * n];
	double *XTP = new double[Plength * l];//X^T P
	// The order is P = [w_i, g_i, w_{i+1}, g_{i+1} ...],
	// Each column represents a possible direction to update, stored in a 1d array but treated as an n * length matrix as in blas
	double *localw = P + wcounter * n;// The latest iterate (w is the final output)
	double *wx = XTP + wcounter * l; // X^T w
	double *g = P + gcounter * n;//gradient
	double *gx = XTP + gcounter * l;// X^T g
	double *system = new double[Plength * Plength];//The coefficients of the lhs of the linear system
	int sparse;//= 1 when DynamicM < M, then solve a sparse linear system by ignoring entries related to the iterates

	tmp_length = 2;//Current length of P = 2*DynamicM + 2

	memset(w, 0, sizeof(double) * size_t(n));
	memcpy(localw, w, sizeof(double) * size_t(n));
	memset(wx, 0, sizeof(double) * size_t(l));

//Initialization: compute f(w_0), P_0, and X^T P_0
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	fun_obj->Xv(g, gx);
	gnorm = dnrm2_(&n, g, &inc);
	gnorm_at_0 = gnorm;
	double tmp[2];// Compute w^T w and g^T w together using blas
	dgemv_(T, &n, &tmp_length, &one, P, &n, localw, &inc, &zero, tmp, &inc);
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp[wcounter];
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = tmp[gcounter];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;

	int64_t time_st = wall_clock_ns();
//Start iterating
	while (iter < max_iter)
	{
//Check stopping and print information
		gnorm = sqrt(inner_product_matrix[inner_product_entry(gcounter, gcounter)]);//the gradient norm squared is an entry in the inner product matrix

		int64_t timer_ed = wall_clock_ns();
		tt += wall_time_diff(timer_ed, time_st);
		time_st = wall_clock_ns();
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm_at_0)
			break;

//Start linear system construction
		sparse = (iter <= M);// When DynamicM < M, solve a sparse linear system by ignoring entries related to the iterates
		tmp_length -= (DynamicM + 1) * sparse;//Reduced system size

//LHS
		fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);//Get D_w, possibly sparse
		if (index_length == 0)//Dense D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse);//Compute the matrix in the linear system
		else//Sparse D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse, index, index_length);


//Now compute the rhs vector in the linear system
		for (i=0;i<tmp_length;i++)
			linear[i] = -inner_product_matrix[inner_product_entry(i + (i + 1) * sparse, gcounter)];//The rhs is -P^T g

//Compute t by solving the linear system
		err = solve_system(system, linear, tmp_length, tmp_length, t);

		if (sparse)//Move the coefficients to the right place for operations with P and XTP
		{
			for (i=tmp_length;i>0;i--)
			{
				t[2 * i - 1] = t[i - 1];
				t[2 * i - 2] = 0;
			}
			tmp_length = 2 * DynamicM + 2;
		}

//Use t to construct information needed for line search
		gTp = 0;
		memset(pTptmp, 0, sizeof(double) * Plength);
		dspmv_(U, &tmp_length, &one, inner_product_matrix, t, &inc, &zero, pTptmp, &inc);//P^T P t
		pTp = ddot_(&tmp_length, t, &inc, pTptmp, &inc);
		gTp = pTptmp[gcounter];
		double wTp = pTptmp[wcounter];

		if (gTp > 0)// If the update direction is nondescent: reverse the direction
		{
			gTp = -gTp;
			wTp = -wTp;
			dscal_(&tmp_length, &minusone, t, &inc);
		}

//If g and p are almost orthogonal, then do gradient descent instead
		if (fabs(gTp) < inner_product_matrix[inner_product_entry(gcounter, gcounter)] * 1e-12)
		{
			memset(t, 0, sizeof(double) * Plength);
			t[gcounter] = -1;
			pTp = inner_product_matrix[inner_product_entry(gcounter, gcounter)];
			gTp = -pTp;
			wTp  = -inner_product_matrix[inner_product_entry(wcounter, gcounter)];
			memset(step, 0, sizeof(double) * n);
			memset(step_x, 0, sizeof(double) * l);
			daxpy_(&n, &minusone, g, &inc, step, &inc);
			daxpy_(&l, &minusone, gx, &inc, step_x, &inc);
		}
		else//Otherwise, compute Pt and (X^T P) t
		{
			dgemv_(N, &n, &tmp_length, &one, P, &n, t, &inc, &zero, step, &inc);
			dgemv_(N, &l, &tmp_length, &one, XTP, &l, t, &inc, &zero, step_x, &inc);
		}

//Line search
		double oldf = f;
		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		pTp = ddot_(&n, step, &inc, step, &inc);
		wTp = ddot_(&n, localw, &inc, step, &inc);
		time_st = wall_clock_ns();
		step_size = fun_obj->line_search(step, gTp, pTp, wTp, &f, eta, 1.0, step_x);

		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}

//Update w, P, X^T P, and P^T P
		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);
		tmp_length = 2 * DynamicM + 2;

		//Pointer to the right place of w
		//Circular storage to avoid moving memory
		wcounter = (wcounter + 2) % Plength;
		gcounter = wcounter + 1;

//First update w, X^T w and f(w)
		if (step_size == 1)//Separated operation when stepsize = 1 for better numerical stability and efficiency
		{
			double *tmp_ptr = P + wcounter * n;
			for (i=0;i<n;i++)
				tmp_ptr[i] = localw[i] + step[i];
			localw = P + wcounter * n;
		}
		else
		{
			memcpy(P + wcounter * n, localw, sizeof(double) * n);
			localw = P + wcounter * n;
			daxpy_(&n, &step_size, step, &inc, localw, &inc);
		}
		wx = XTP + wcounter * l;
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		f = fun_obj->fun(localw);//New objective
		time_st = wall_clock_ns();
		double actred = oldf - f;

		//numerical error handling
		if (fabs(actred) < 0)
		{
			info("WARNING: actred < 0\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}

		fun_obj->get_wx(wx);

//Update P and X^T P
		g = P + gcounter * n;
		gx = XTP + gcounter * l;
		fun_obj->grad(localw, g);
		fun_obj->Xv(g, gx);

//Update P^T P
		update_inner_products(inner_product_matrix, DynamicM, previous_M, P, t, step_size);
	}
	memcpy(w, localw, sizeof(double) * n);

	delete[] pTptmp;
	delete[] t;
	delete[] system;
	delete[] XTP;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
}

void LCOMM_BFGS::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse)
{
	double one = 1.0;
	char U[] = "U";
	char T[] = "T";

	int i,j;
	for (i=0;i<tmp_length;i++)
	{
		int counteri = i * l;
		int tmp_counter = (i + (i + 1) * sparse) * l;//Skip the columns for w_i if sparse = 1
		//Fill in the entries of sqrt(D) X^T P corresponds to the i-th column of X^T P
		for (j=0;j<l;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + j];
	}

	//Compute the part of the linear system from P^T P
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i + (i + 1) * sparse, j + (j + 1) * sparse)];

	dsyrk_(U, T, &tmp_length, &l, &one, sqrtD_XTP, &l, &one, system, &tmp_length);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];//Update the lower triangular part by copy-pasting
}

//When D_w is sparse
void LCOMM_BFGS::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length)
{
	double one = 1.0;
	char U[] = "U";
	char T[] = "T";

	int i,j;
	for (i=0;i<tmp_length;i++)
	{
		int counteri = i * index_length;
		int tmp_counter = (i + (i + 1) * sparse) * l;
		//Fill in the entries of sqrt(D) X^T P corresponds to the i-th column of X^T P, ignore columns from X^T P corresponding to zero D_w entries
		for (j=0;j<index_length;j++)
			sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + index[j]];
	}

	//Compute the part of the linear system from P^T P
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i + (i + 1) * sparse, j + (j + 1) * sparse)];

	dsyrk_(U, T, &tmp_length, &index_length, &one, sqrtD_XTP, &index_length, &one, system, &tmp_length);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part, and note that sqrtD_XTP has only index_length rows

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];
}


void LCOMM_BFGS::update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size)
{
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";
	int inc = 1;

	int i;
	int prev_length = 2 * previous_M + 2;//#columns of P in last iteration
	double *PTg = new double[tmp_length];
	double *PTw = new double[tmp_length];
	dgemv_(T, &n, &tmp_length, &one, P, &n, P + gcounter * n, &inc, &zero, PTg, &inc);//PTg = P^T g

	int prev_w = wcounter - 2;//Place of last iterate
	if (wcounter == 0)
		prev_w = 2 * DynamicM;//circular storage

	if (step_size != 1)
		dscal_(&prev_length, &step_size, t, &inc);//Incorporate stepsize into t to reduce calculation
	t[prev_w] += 1;//Now w_new = w_old + step_size*p = Pt

	dspmv_(U, &prev_length, &one, inner_product_matrix, t, &inc, &zero, PTw, &inc);
	PTw[wcounter] = ddot_(&prev_length, PTw, &inc, t, &inc);

	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, wcounter)] = PTw[i];
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];


	delete[] PTg;
	delete[] PTw;
}

LCOMM_DIAG::LCOMM_DIAG(const function *fun_obj, int l, double eps, int m, double eta, int max_iter): LCOMM_BFGS(fun_obj, l, eps, m, eta, max_iter)
{
	this->M = m;//M is #historical iterations
	Plength = 3 * M + 3;//3 vectors per iteration in P
	this->n = this->fun_obj->get_nr_variable();
}

LCOMM_DIAG::~LCOMM_DIAG()
{
}

//Algorithm 3 in the paper. Here A = I and b = 0, so many operations are simplified
void LCOMM_DIAG::solve(double *w)
{
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int iter = 0, inc = 1;
	int two = 2;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";

	sqrtD_XTP = new double[Plength * l];
	int i;
	double f;
	double gnorm;
	double step_size = 0;
	double gTp = 0;
	double err = 0;
	double pTp;
	double gnorm_at_0;
	double tt = 0;//Timer
	int *index = NULL; //Index set, used when D_w and u_w are sparse
	int index_length = 0;//the size of the index set. Set to 0 when the vectors are dense

	int DynamicM = 0;//min(M,iter)
	int previous_M = 0;//DynamicM at last iteration

	wcounter = 0;//the column of P that stores w
	gcounter = 1;//The column of P that stores the gradient
	dcounter = 2;//The column of P that stores the current D^{-1} g
	
	double* linear = new double[Plength];//The rhs of the linear system
	double* pTptmp = new double[Plength];//A temp vector for computing p^T p using the inner product matrix
	double *sqrt_C_times_D = new double[l];//sqrt(D_w)
	double *inner_product_matrix = new double[Plength * Plength];//symmetric matrix, so only need to store the upper triangular part
	double *step = new double[n];//This is p
	double *step_x = new double[l];//This is X^T p
	double* t = new double[Plength];// Stores the current solution to the linear system
	double *P = new double[Plength * n];
	double *XTP = new double[Plength * l];//X^T P
	// The order is P = [w_i, g_i, D_i^{-1} g_i, w_{i+1}, g_{i+1}, D_{i+1}^{-1} g_{i+1} ...,
	double *localw = P + wcounter * n;// The latest iterate (w is the final output)
	double *wx = XTP + wcounter * l; // X^T w
	double *g = P + gcounter * n;//gradient
	double *gx = XTP + gcounter * l;// X^T g
	double *D_inverse_g = P + dcounter * n;
	double *D_inverse_g_x = XTP + dcounter * l;
	double *system = new double[Plength * Plength];//The coefficients of the lhs of the linear system
	int sparse;//= 1 when DynamicM < M, then solve a sparse linear system by ignoring entries related to the iterates

	tmp_length = 3;//Current length of P

	memset(w, 0, sizeof(double) * size_t(n));
	memcpy(localw, w, sizeof(double) * size_t(n));
	memset(wx, 0, sizeof(double) * size_t(l));

//Initialization: compute f(w_0), P_0, and X^T P_0
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	gnorm = dnrm2_(&n, g, &inc);
	gnorm_at_0 = gnorm;// For stopping condition
	fun_obj->get_diagH(D_inverse_g);//First store D
	for (i=0;i<n;i++)
		D_inverse_g[i] = g[i] / D_inverse_g[i];//Change to g/D
	fun_obj->Xv(g, gx, 2);//D_inverse_g is right after g, so compute together

	double tmp[5];//Compute the initial inner product entries using blas so need a buffer

	dgemv_(T, &n, &tmp_length, &one, P, &n, P + dcounter * n, &inc, &zero, tmp, &inc);//P^T (inv(D)*g) in the first 3 entries of tmp
	dgemv_(T, &n, &two, &one, P, &n, P + wcounter * n, &inc, &zero, tmp + 3, &inc);//P^T w in the last 2 entries (ignore the repeated computation for w and D^{-1} g

	inner_product_matrix[inner_product_entry(dcounter, wcounter)] = tmp[wcounter];
	inner_product_matrix[inner_product_entry(dcounter, gcounter)] = tmp[gcounter];
	inner_product_matrix[inner_product_entry(dcounter, dcounter)] = tmp[dcounter];
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp[wcounter + 3];
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = tmp[gcounter + 3];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;


	int64_t time_st = wall_clock_ns();
//Start iterating
	while (iter < max_iter)
	{
//Check stopping and print information
		gnorm = sqrt(inner_product_matrix[inner_product_entry(gcounter, gcounter)]);//the gradient norm squared is an entry in the inner product matrix

		int64_t timer_ed = wall_clock_ns();
		tt += wall_time_diff(timer_ed, time_st);
		time_st = wall_clock_ns();
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm_at_0)
			break;

//Start linear system construction
		sparse = (iter <= M);// When DynamicM < M, solve a sparse linear system by ignoring entries related to the iterates
		tmp_length -= (DynamicM + 1) * sparse;//Reduced system size

//LHS
		fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);//Get D_w, possibly sparse
		if (index_length == 0)//Dense D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse);//Compute the matrix in the linear system
		else//Sparse D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse, index, index_length);

//Now compute the rhs vector in the linear system
		if (sparse)
			for (i=0;i<tmp_length / 2;i++)
			{
				linear[2 * i] = -inner_product_matrix[inner_product_entry(3 * i + 1, gcounter)];
				linear[2 * i + 1] = -inner_product_matrix[inner_product_entry(3 * i + 2, gcounter)];
			}
		else
			for (i=0;i<tmp_length;i++)
				linear[i] = -inner_product_matrix[inner_product_entry(i, gcounter)];

//Compute t by solving the linear system
		err = solve_system(system, linear, tmp_length, tmp_length, t);

		if (sparse)//Move the coefficients to the right place for operations with P and XTP
		{
			for (i=tmp_length / 2;i>0;i--)
			{
				t[3 * i - 1] = t[2 * i - 1];
				t[3 * i - 2] = t[2 * i - 2];
				t[3 * i - 3] = 0;
			}
			tmp_length = 3 * DynamicM + 3;
		}

//Use t to construct information needed for line search
		gTp = 0;
		memset(pTptmp, 0, sizeof(double) * Plength);
		dspmv_(U, &tmp_length, &one, inner_product_matrix, t, &inc, &zero, pTptmp, &inc);//P^T P t
		pTp = ddot_(&tmp_length, t, &inc, pTptmp, &inc);
		gTp = pTptmp[gcounter];
		double wTp = pTptmp[wcounter];

		if (gTp > 0)// If the update direction is nondescent: reverse the direction
		{
			gTp = -gTp;
			wTp = -wTp;
			dscal_(&tmp_length, &minusone, t, &inc);
		}
		
//If g and p are almost orthogonal, then do gradient descent instead
		if (fabs(gTp) < inner_product_matrix[inner_product_entry(gcounter, gcounter)] * 1e-12)
		{
			memset(t, 0, sizeof(double) * Plength);
			t[gcounter] = -1;
			pTp = inner_product_matrix[inner_product_entry(gcounter, gcounter)];
			gTp = -pTp;
			wTp  = -inner_product_matrix[inner_product_entry(wcounter, gcounter)];
			memset(step, 0, sizeof(double) * n);
			memset(step_x, 0, sizeof(double) * l);
			daxpy_(&n, &minusone, g, &inc, step, &inc);
			daxpy_(&l, &minusone, gx, &inc, step_x, &inc);
		}
		else//Otherwise, compute Pt and (X^T P) t
		{
			dgemv_(N, &n, &tmp_length, &one, P, &n, t, &inc, &zero, step, &inc);
			dgemv_(N, &l, &tmp_length, &one, XTP, &l, t, &inc, &zero, step_x, &inc);
		}

//Line search
		double oldf = f;
		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		pTp = ddot_(&n, step, &inc, step, &inc);
		wTp = ddot_(&n, localw, &inc, step, &inc);
		time_st = wall_clock_ns();
		step_size = fun_obj->line_search(step, gTp, pTp, wTp, &f, eta, 1.0, step_x);

		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}

//Update w, P, X^T P, and P^T P
		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);
		tmp_length = 3 * DynamicM + 3;

		//Pointer to the right place of w
		//Circular storage to avoid moving memory
		wcounter = (wcounter + 3) % Plength;
		gcounter = wcounter + 1;
		dcounter = wcounter + 2;

//First update w, X^T w and f(w)
		if (step_size == 1)//Separated operation when stepsize = 1 for better numerical stability and efficiency
		{
			double *tmp_ptr = P + wcounter * n;
			for (i=0;i<n;i++)
				tmp_ptr[i] = localw[i] + step[i];
			localw = P + wcounter * n;
		}
		else
		{
			memcpy(P + wcounter * n, localw, sizeof(double) * n);
			localw = P + wcounter * n;
			daxpy_(&n, &step_size, step, &inc, localw, &inc);
		}
		wx = XTP + wcounter * l;

		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		f = fun_obj->fun(localw);
		time_st = wall_clock_ns();
		double actred = oldf - f;
		//
		//numerical error handling
		if (fabs(actred) < 0)
		{
			info("WARNING: actred < 0\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}
		if (fabs(actred) <= 1.0e-20*fabs(f))
		{
			info("WARNING: actred too small\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}

		fun_obj->get_wx(wx);

//Update P and X^T P
		g = P + gcounter * n;
		gx = XTP + gcounter * l;
		fun_obj->grad(localw, g);

		D_inverse_g = P + dcounter * n;
		D_inverse_g_x = XTP + dcounter * l;
		fun_obj->get_diagH(D_inverse_g);//First store D
		for (i=0;i<n;i++)
			D_inverse_g[i] = g[i] / D_inverse_g[i];//Change to g/D
		fun_obj->Xv(g, gx, 2);//D_inverse_g is right after g, so compute together

//Update P^T P
		update_inner_products(inner_product_matrix, DynamicM, previous_M, P, t, step_size);
	}
	memcpy(w, localw, sizeof(double) * n);

	delete[] pTptmp;
	delete[] t;
	delete[] system;
	delete[] XTP;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
}

void LCOMM_DIAG::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse)
{
	double one = 1.0;
	char U[] = "U";
	char T[] = "T";

	int i,j;
	if (sparse)
	{
		for (i=0;i<tmp_length/2;i++)//Only skip columns = 3j for the iterates
		{
			int counteri = 2 * i * l;
			int tmp_counter = (3 * i + 1) * l;
			for (j=0;j<l;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + j];
			counteri += l;
			tmp_counter = (3 * i + 2) * l;
			for (j=0;j<l;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + j];

			//The following 2 fors compute the part of the linear system from P^T P
			for (j=0;j < i;j++)
			{
				system[2 * i * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 1)];
				system[2 * i * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 2)];
			}
			system[2 * i * tmp_length + 2 * i] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * i + 1)];
			for (j=0;j <= i;j++)
			{
				system[(2 * i + 1) * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 1)];
				system[(2 * i + 1) * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 2)];
			}
		}
	}
	else
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * l;
			for (j=0;j<l;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[counteri + j];
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i,j)];
		}

	dsyrk_(U, T, &tmp_length, &l, &one, sqrtD_XTP, &l, &one, system, &tmp_length);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];//Update the lower triangular part by copy-pasting
}

//When D_w is sparse
void LCOMM_DIAG::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length)
{
	double one = 1.0;
	char U[] = "U";
	char T[] = "T";

	int i,j;
	if (sparse)
	{
		for (i=0;i<tmp_length/2;i++)
		{
			int counteri = 2 * i * index_length;
			int tmp_counter = (3 * i + 1) * l;
			for (j=0;j<index_length;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + index[j]];
			counteri += index_length;
			tmp_counter += l;
			for (j=0;j<l;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + index[j]];
			for (j=0;j < i;j++)
			{
				system[2 * i * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 1)];
				system[2 * i * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 2)];
			}
			system[2 * i * tmp_length + 2 * i] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * i + 1)];

			for (j=0;j <= i;j++)
			{
				system[(2 * i + 1) * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 1)];
				system[(2 * i + 1) * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 2)];
			}
		}
	}
	else
		for (i=0;i<tmp_length;i++)
		{
			int counteri = i * index_length;
			int tmp_counter = i * l;
			for (j=0;j<index_length;j++)
				sqrtD_XTP[counteri + j] = sqrt_C_times_D[j] * XTP[tmp_counter + index[j]];
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i,j)];
		}

	dsyrk_(U, T, &tmp_length, &index_length, &one, sqrtD_XTP, &index_length, &one, system, &tmp_length);

	for (i=0;i<tmp_length;i++)
		for (j=i+1;j<tmp_length;j++)
			system[i * tmp_length + j] = system[j * tmp_length + i];
}


void LCOMM_DIAG::update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size)
{
	int inc = 1;
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";

	int i;
	int prev_length = 3 * previous_M + 3;//#columns of P in last iteration
	double *PTg = new double[tmp_length * 2];
	double *PTd = PTg + tmp_length;//d stands for inv(D)*g
	double *PTw = new double[tmp_length];
	dgemv_(T, &n, &tmp_length, &one, P, &n, P + gcounter * n, &inc, &zero, PTg, &inc);//P^T g
	dgemv_(T, &n, &tmp_length, &one, P, &n, P + dcounter * n, &inc, &zero, PTd, &inc);//P^T d

	int prev_w = wcounter - 3;//Place of last iterate
	if (wcounter == 0)
		prev_w = 3 * DynamicM;//circular storage

	if (step_size != 1)
		dscal_(&prev_length, &step_size, t, &inc);//Incorporate stepsize into t to reduce calculation
	t[prev_w] += 1;//Now w_new = w_old + step_size*p = Pt

	dspmv_(U, &prev_length, &one, inner_product_matrix, t, &inc, &zero, PTw, &inc);
	PTw[wcounter] = ddot_(&prev_length, PTw, &inc, t, &inc);

	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, wcounter)] = PTw[i];
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, dcounter)] = PTd[i];

	delete[] PTw;
	delete[] PTg;
}

//The following is an implementation of accelerated gradient
AG::AG(const function *fun_obj, double eps, int linesearch, int max_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->linesearch = linesearch;
}

void AG::solve(double *w)
{
	int n = fun_obj->get_nr_variable();
	double sigma = 1;
	int inc = 1;
	double beta = 0.5;
	double beta_square = 0.25;
	double theta = 1;
	double theta_pre = 1;
	double t = 1;
	int i = 1;
	double *x = new double[n];
	double *x_pre = new double[n];
	double *v = new double[n];
	double *y = new double[n];
	double *gy = new double[n];
	double f;
	memset(x_pre, 0, sizeof(double) * n);
	memcpy(v, x_pre, sizeof(double) * n);
	double *g = new double[n];
	int search = 1;
	f = fun_obj->fun(x_pre);
	fun_obj->grad(x_pre, g);
	double t_pre = 1.0 / fun_obj->compute_rho();

	double gnorm_at_0 = dnrm2_(&n, g, &inc);
	double gnorm = gnorm_at_0;

	if (gnorm <= eps*gnorm_at_0)
		search = 0;

	double tt = 0;
	info("iter %d |g| %5.3e f %15.20e time %.4e\n", 0, gnorm_at_0, f, tt);
	int64_t time_st = wall_clock_ns();
	while (i <= max_iter && search)
	{
		if (linesearch)
			t = t_pre / beta_square;
		else
			t = t_pre;
		int counter = 0;
		while (true)
		{
			if (linesearch)
			{
				t *= beta;
				double a = 1;
				double tmp = theta_pre * theta_pre * t / t_pre;
				double b = - sigma * t + tmp;
				double c = - tmp;
				theta = (- b + sqrt((b * b) - 4 * a * c)) / (2 * a);
			}
			else
			{
				theta = sqrt(sigma * t);
				theta_pre = theta;
			}
			double theta2 = theta / (1 + (theta / theta_pre / theta_pre) * t_pre * sigma);
			double coeff = 1 - theta2;

			for (int j=0;j<n;j++)
				y[j] = coeff * x_pre[j] + theta2 * v[j];

			double fy = fun_obj->fun(y);
			fun_obj->grad(y, gy);
			counter++;

			// compute proximal mapping
			double inner_product = 0;
			for (int j=0;j<n;j++)
				x[j] = y[j] - t * gy[j];
			if (linesearch)
			{
				inner_product = - ddot_(&n, gy, &inc, gy, &inc) * t * 0.5;
				f = fun_obj->fun(x);
				if (f <= fy + inner_product)
					break;
			}
			else
				break;
		}
		for (int j = 0; j < n; j++)
			v[j] = x_pre[j] + (x[j] - x_pre[j]) / theta;

		// update x, theta_pre and t_pre for next iteration
		delete[] x_pre;
		x_pre = x;
		theta_pre = theta;
		t_pre = t;
		int64_t iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st);
		if (linesearch == 0)
			f = fun_obj->fun(x_pre);
		fun_obj->grad(x_pre, g);
		gnorm = dnrm2_(&n, g, &inc);
		info("iter %d |g| %5.3e f %15.20e time %.4e\n", i, gnorm, f, tt);
		if (gnorm <= eps*gnorm_at_0)
			break;
		time_st = wall_clock_ns();
		i++;
		x = new double[n];
	}
	delete[] y;
	delete[] v;
	delete[] gy;
	delete[] g;
	memcpy(w,x,sizeof(double) * n);
	delete[] x;
}
